"""
Copyright  2016-2019 Maël Azimi <m.a@moul.re>

Silkaj is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Silkaj is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with Silkaj. If not, see <https://www.gnu.org/licenses/>.
"""

from gdon.silkaj.tools import message_exit
#from click import command, option, pass_context, confirm
from getpass import getpass
from pathlib import Path
from re import compile, search
from duniterpy.key import SigningKey
from duniterpy.key.scrypt_params import ScryptParams


#@pass_context
def auth_method(ctx):
    if ctx.obj["AUTH_SEED"]:
        return auth_by_seed()
    if ctx.obj["AUTH_FILE"]:
        return auth_by_auth_file()
    if ctx.obj["AUTH_WIF"]:
        return auth_by_wif()
    else:
        return auth_by_scrypt()


#@command("authfile", help="Generate authentication file")
#@option("--file", default="authfile", show_default=True, help="Path file")
def generate_auth_file(file):
    key = auth_method()
    authfile = Path(file)
    if authfile.is_file():
        confirm(
            "Would you like to erase "
            + file
            + " by an authfile corresponding to following pubkey `"
            + key.pubkey
            + "`?",
            abort=True,
        )
    key.save_seedhex_file(file)
    print(
        "Authentication file 'authfile' generated and stored in current\
 folder for following public key:",
        key.pubkey,
    )


#@pass_context
def auth_by_auth_file(ctx):
    file = ctx.obj["AUTH_FILE_PATH"]
    authfile = Path(file)
    if not authfile.is_file():
        message_exit('Error: the file "' + file + '" does not exist')
    filetxt = authfile.open("r").read()
    regex_seed = compile("^[0-9a-fA-F]{64}$")
    regex_gannonce = compile(
        "^pub: [1-9A-HJ-NP-Za-km-z]{43,44}\nsec: [1-9A-HJ-NP-Za-km-z]{88,90}.*$"
    )
    # Seed Format
    if search(regex_seed, filetxt):
        return SigningKey.from_seedhex_file(file)
    # gannonce.duniter.org Format
    elif search(regex_gannonce, filetxt):
        return SigningKey.from_pubsec_file(file)
    else:
        message_exit("Error: the format of the file is invalid")


def auth_by_seed():
    seedhex = getpass("Please enter your seed on hex format: ")
    try:
        return SigningKey.from_seedhex(seedhex)
    except Exception as error:
        message_exit(error)

"""
#@pass_context
def auth_by_scrypt(ctx):
    salt = getpass("Please enter your Scrypt Salt (Secret identifier): ")
    password = getpass("Please enter your Scrypt password (masked): ")

    if ctx.obj["AUTH_SCRYPT_PARAMS"]:
        n, r, p = ctx.obj["AUTH_SCRYPT_PARAMS"].split(",")

        if n.isnumeric() and r.isnumeric() and p.isnumeric():
            n, r, p = int(n), int(r), int(p)
            if n <= 0 or n > 65536 or r <= 0 or r > 512 or p <= 0 or p > 32:
                message_exit("Error: the values of Scrypt parameters are not good")
            scrypt_params = ScryptParams(n, r, p)
        else:
            message_exit("one of n, r or p is not a number")
    else:
        scrypt_params = None

    try:
        return SigningKey.from_credentials(salt, password, scrypt_params)
    except ValueError as error:
        message_exit(error)
"""

def auth_by_scrypt(ctx):
    salt = getpass("Please enter your Scrypt Salt (Secret identifier): ")
    password = getpass("Please enter your Scrypt password (masked): ")

#    if ctx.obj["AUTH_SCRYPT_PARAMS"]:
#        n, r, p = ctx.obj["AUTH_SCRYPT_PARAMS"].split(",")

#        if n.isnumeric() and r.isnumeric() and p.isnumeric():
#            n, r, p = int(n), int(r), int(p)
#            if n <= 0 or n > 65536 or r <= 0 or r > 512 or p <= 0 or p > 32:
#                message_exit("Error: the values of Scrypt parameters are not good")
#            scrypt_params = ScryptParams(n, r, p)
#        else:
#            message_exit("one of n, r or p is not a number")
#    else:
#        scrypt_params = None

    scrypt_params = None
    try:
        return SigningKey.from_credentials(salt, password, scrypt_params)
    except ValueError as error:
        message_exit(error)

def auth_by_scrypt_with_IDs(salt, password):
    scrypt_params = None
    try:
        return SigningKey.from_credentials(salt, password, scrypt_params)
    except ValueError as error:
        message_exit(error)



def auth_by_wif():
    wif_hex = getpass("Enter your WIF or Encrypted WIF address (masked): ")
    password = getpass(
        "(Leave empty in case WIF format) Enter the Encrypted WIF password (masked): "
    )
    try:
        return SigningKey.from_wif_or_ewif_hex(wif_hex, password)
    except Exception as error:
        message_exit(error)



import datetime
import os
import json
import re

from gdon.constants import DEFAULT_FOLDER, WALLETS_FOLDER, BACKUP_FOLDER, PDF_BACKUP_FOLDER, PDF_EXTENSION, SIGNED_EXTENSION, CONF_FILE, HTTP_PORT
from gdon.silkaj.constants import G1_DEFAULT_ENDPOINT

def check_date(given_date):
    """
    check if a given peremption date is correctly formatted.
    if in the past, asks for user validation.
    """
    infos_date = given_date.split("/")
    if len(infos_date) != 3 or len(infos_date[2]) != 4 or len(infos_date[0]) != 2 or len(infos_date[1]) != 2:
        print ("Erreur : le champs <date> doit être de forme JJ/MM/YYYY")
        exit
    today = datetime.datetime.now()
    date = datetime.datetime(int(infos_date[2]), int(infos_date[1]), int(infos_date[0]))
    if today > date:
        while True:
            answer = input("La date donnée est passée. Voulez-vous continuer (O/n) ? : ")
            if answer == "n" or answer == "N":
                print ("Veuillez recommencer.")
                exit(0)
            elif answer == "o" or answer == "O" or answer == "":
                break
            else:
                print("La réponse doit être O ou n.")
    return datetime.date(int(infos_date[2]), int(infos_date[1]), int(infos_date[0]))


class CheckConfig():
    """
    check that the config exists. If not, create relevant folders.
    """
    def __init__(self):

        self.config_file = os.path.expanduser(CONF_FILE)
        if not os.path.exists(os.path.dirname(self.config_file)):
            os.mkdir(os.path.dirname(self.config_file))
        if os.path.exists(self.config_file):
            with open(self.config_file, "r") as f:
                conf = json.loads(f.read())
                folder = conf["folder"]
        else :
            folder = os.path.expanduser(DEFAULT_FOLDER)

        self.folder = os.path.expanduser(folder)
        self.wallets_folder = os.path.expanduser(folder + WALLETS_FOLDER)
        self.backup_folder = os.path.expanduser(folder + BACKUP_FOLDER)

    def __call__(self):
        return self

    def check(self, node=G1_DEFAULT_ENDPOINT):
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)
        if not os.path.exists(self.wallets_folder):
            os.makedirs(self.wallets_folder)
        if not os.path.exists(self.backup_folder):
            os.makedirs(self.backup_folder)
        if not os.path.exists(self.config_file):
            f = open(self.config_file, "x")
            f.close()
            self.write_conf(node[0], node[1], self.folder)

    def get_conf(self):
        if not os.path.exists(self.config_file):
            self.check()
        with open(self.config_file, "r") as f:
            return json.loads(f.read())

    def display_conf(self):
        if not os.path.exists(self.config_file):
            self.check()
        with open(self.config_file, "r") as f:
            conf = json.loads(f.read())
            print ("\nConfiguration :\nnode : {0}\nport : {1}\ndossier : {2}\n".format(conf["node"], conf["port"], conf["folder"]))


    def check_conf(self, node: str, port: int, folder:str):
        conf = self.get_conf()
        if node == None:
            node = conf["node"]
            if port == None:
                port = conf["port"]

        if folder != None:
            folder = os.path.abspath(folder) + "/"
        elif folder == None:
            folder = conf["folder"]

        self.write_conf(node, port, folder)
        self.check()

    def write_conf(self, given_node, port, folder):
        conf = dict()
        conf["node"] = str(given_node)
        conf["port"] = str(int(port))
        conf["folder"] = str(folder)
        with open(self.config_file, "w") as f:
            f.write(json.dumps(conf, indent=4))
        print("Configuration enregistrée : \n node : {0}\n port : {1}\n dossier : {2}".format(given_node, str(port), folder))



def backup_data(data, name):
    """
    Backups the signed files to .backup
    If there is a corresponding pdf file in the conf folder, move it to the .backup/year/pdf folder.
    Takes as inputs the full path and the name of the signed document.
    """

    conf = CheckConfig()
    conf.check()

    # check that the backup folder exists
    year = str(datetime.datetime.now().year)
    backup_folder = os.path.join(conf.backup_folder, year)
    if not os.path.exists(backup_folder):
        os.makedirs(backup_folder)

    # create the backup path and move the file
    backup = os.path.join(backup_folder, name)
    os.rename(data, backup)
    print("{0} --> {1}/".format(name, backup_folder))

    # check if there is a pdf file in the default wallets folder, and move it to the backups pdf folder
    if re.search(SIGNED_EXTENSION, name):
        pdf_file = name.replace(SIGNED_EXTENSION, PDF_EXTENSION)
        pdf_folder = os.path.join(backup_folder, PDF_BACKUP_FOLDER)
        pdf_orig = os.path.expanduser(conf.folder + pdf_file)
        if os.path.isfile(pdf_orig):
            # check for pdf_folder and create it
            if not os.path.exists(pdf_folder):
                os.makedirs(pdf_folder)
            os.rename(pdf_orig, pdf_folder + pdf_file)
        print("{0} -->  {1}".format(pdf_file, pdf_folder))




"""
Copyright  2016-2019 Maël Azimi <m.a@moul.re>

Silkaj is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Silkaj is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with utils.silkaj. If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import unicode_literals
from ipaddress import ip_address
import socket
import logging
from sys import exit, stderr
#from click import pass_context
from asyncio import sleep
from duniterpy.api.client import Client
from duniterpy.api.bma import blockchain, network

from gdon.silkaj.constants import (
    G1_DEFAULT_ENDPOINT,
    G1_TEST_DEFAULT_ENDPOINT,
    CONNECTION_TIMEOUT,
    ASYNC_SLEEP,
)

from gdon.utils import CheckConfig 
#from gui.parameters import ClientInstance 


async def discover_peers(discover):
    """
    From first node, discover his known nodes.
    Remove from know nodes if nodes are down.
    If discover option: scan all network to know all nodes.
        display percentage discovering.
    """
    client = ClientInstance().client
    endpoints = await get_peers_among_leaves(client)
    if discover:
        print("Discovering network")
    for i, endpoint in enumerate(endpoints):
        if discover:
            print("{0:.0f}%".format(i / len(endpoints) * 100))
        if best_endpoint_address(endpoint, False) is None:
            endpoints.remove(endpoint)
        elif discover:
            endpoints = await recursive_discovering(endpoints, endpoint)
    return endpoints


async def recursive_discovering(endpoints, endpoint):
    """
    Discover recursively new nodes.
    If new node found add it and try to found new node from his known nodes.
    """
    api = generate_duniterpy_endpoint_format(endpoint)
    sub_client = Client(api)
    news = await get_peers_among_leaves(sub_client)
    await sub_client.close()
    for new in news:
        if best_endpoint_address(new, False) is not None and new not in endpoints:
            endpoints.append(new)
            await recursive_discovering(endpoints, new)
    return endpoints


async def get_peers_among_leaves(client):
    """
    Browse among leaves of peers to retrieve the other peers’ endpoints
    """
    leaves = await client(network.peers, leaves=True)
    peers = list()
    for leaf in leaves["leaves"]:
        await sleep(ASYNC_SLEEP + 0.05)
        leaf_response = await client(network.peers, leaf=leaf)
        peers.append(leaf_response["leaf"]["value"])
    return parse_endpoints(peers)


def parse_endpoints(rep):
    """
    endpoints[]{"domain", "ip4", "ip6", "pubkey"}
    list of dictionaries
    reps: raw endpoints
    """
    i, j, endpoints = 0, 0, []
    while i < len(rep):
        if rep[i]["status"] == "UP":
            while j < len(rep[i]["endpoints"]):
                ep = parse_endpoint(rep[i]["endpoints"][j])
                j += 1
                if ep is None:
                    break
                ep["pubkey"] = rep[i]["pubkey"]
                endpoints.append(ep)
        i += 1
        j = 0
    return endpoints


def generate_duniterpy_endpoint_format(ep):
    api = "BASIC_MERKLED_API " if ep["port"] != "443" else "BMAS "
    api += ep.get("domain") + " " if "domain" in ep else ""
    api += ep.get("ip4") + " " if "ip4" in ep else ""
    api += ep.get("ip6") + " " if "ip6" in ep else ""
    api += ep.get("port")
    return api


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return getinstance

@singleton
#class EndPoint(object):
class EndPoint():
#    @pass_context
#    def __init__(ctx, self):
    def __init__(self):
        ep = CheckConfig().get_conf()
        api = "BMAS" if ep["port"] == "443" else "BASIC_MERKLED_API"
        self.BMA_ENDPOINT = " ".join([api, ep["node"], ep["port"]])

@singleton
class ClientInstance(object):
    def __init__(self):
#        self.client = Client(EndPoint().BMA_ENDPOINT)
        self.client = Client(EndPoint().BMA_ENDPOINT)

def parse_endpoint(rep):
    """
    rep: raw endpoint, sep: split endpoint
    domain, ip4 or ip6 could miss on raw endpoint
    """
    ep, sep = {}, rep.split(" ")
    if sep[0] == "BASIC_MERKLED_API":
        if check_port(sep[-1]):
            ep["port"] = sep[-1]
        if (
            len(sep) == 5
            and check_ip(sep[1]) == 0
            and check_ip(sep[2]) == 4
            and check_ip(sep[3]) == 6
        ):
            ep["domain"], ep["ip4"], ep["ip6"] = sep[1], sep[2], sep[3]
        if len(sep) == 4:
            ep = endpoint_type(sep[1], ep)
            ep = endpoint_type(sep[2], ep)
        if len(sep) == 3:
            ep = endpoint_type(sep[1], ep)
        return ep
    else:
        return None


def endpoint_type(sep, ep):
    typ = check_ip(sep)
    if typ == 0:
        ep["domain"] = sep
    elif typ == 4:
        ep["ip4"] = sep
    elif typ == 6:
        ep["ip6"] = sep
    return ep


def check_ip(address):
    try:
        return ip_address(address).version
    except:
        return 0


def best_endpoint_address(ep, main):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(CONNECTION_TIMEOUT)
    addresses, port = ("domain", "ip6", "ip4"), int(ep["port"])
    for address in addresses:
        if address in ep:
            try:
                s.connect((ep[address], port))
                s.close()
                return address
            except Exception as e:
                logging.debug(
                    "Connection to endpoint %s (%s) failled (%s)" % (ep, address, e)
                )
    if main:
        print("Wrong node given as argument", file=stderr)
        exit(1)
    return None


def check_port(port):
    try:
        port = int(port)
    except:
        print("Port must be an integer", file=stderr)
        return False
    if port < 0 or port > 65536:
        print("Wrong port number", file=stderr)
        return False
    return True


class HeadBlock(object):
    __instance = None

    def __new__(cls):
        if HeadBlock.__instance is None:
            HeadBlock.__instance = object.__new__(cls)
        return HeadBlock.__instance

    def __init__(self):
        self.head_block = self.get_head()

    async def get_head(self):
        client = ClientInstance().client
        return await client(blockchain.current)



import os
import shutil
import json
import re
from datetime import datetime
from time import sleep

from duniterpy.key import SigningKey, ascii_armor

from gdon.constants import DEFAULT_FOLDER, BACKUP_FOLDER, WALLETS_FOLDER, PDF_BACKUP_FOLDER, SIGNED_EXTENSION, PDF_EXTENSION, COMMENT, MAX_ISSUERS
from gdon.diceware import diceware
from gdon.utils import check_date, backup_data, CheckConfig


from gdon.silkaj.money import get_amount_from_pubkey
from gdon.silkaj.network_tools import ClientInstance
from gdon.silkaj.tx_retrieve import handle_intermediaries_transactions, generate_and_send_transaction, get_list_input_for_transaction
from gdon.silkaj.auth import auth_by_scrypt_with_IDs, auth_by_scrypt
from gdon.silkaj.tools import coroutine, message_exit
from gdon.silkaj.constants import PUBKEY_PATTERN


class Retrieve():
    """
    Retrieve wallets from encrypted files
    """
    def __init__(self, folder=None, _file=None):
        """
        check given values and set self values.
        """

        conf = CheckConfig()
        conf.check()

        if folder == None and _file == None:
             self.folder = conf.wallets_folder
        elif folder != None and _file != None:
            message_exit("Erreur : Retrieve() ne peut recevoir folder ET _file")

        elif folder != None:
            self.folder = os.path.abspath(folder)
            if not os.path.isdir(self.folder):
                message_exit("Erreur : le dossier n'existe pas : {0}".format(self.folder))
        elif _file != None:
            self.file = os.path.abspath(_file)
            if not os.path.isfile(self.file):
                message_exit("Erreur : le  fichier n'existe pas : {0}".format(self.file))

    def open_file(self, file):
        with open(file) as wallets_json:
            return json.load(wallets_json)

    def get_key_list(self, retrieve_wallets):
        """
        create key for each wallet
        """
        issuers_keys = list()
        for wallet in retrieve_wallets:
            key = auth_by_scrypt_with_IDs(wallet['salt'], wallet['password'])
            issuers_keys.append(key)
        return issuers_keys

    def is_date_passed(self, wallets_doc):
        peremption_date = datetime.fromisoformat(wallets_doc["peremption_date"])
        today = datetime.now()
        if not today > peremption_date:
            return False
        return True

    def get_pubkey(self, pubkey):
        """
        get credentials for each creator account
        """
        while True:
            print("\nEntrez les identifiants pour les clef publique : {0}\nSi vous ne savez pas, entrez des identifiants vides.".format(pubkey))
            key = auth_by_scrypt(None)
            if re.search(pubkey, key.pubkey):
                return key
            else:
                answer = input("Mauvais identifiants ! Voulez-vous réessayer ? (o/n) : ")
                if answer == "n":
                    return None


    def sort_folder(self, retrieve_files):
        """
        sorts retrievable files
        """
        # get all files from folder
        file_list = list()
        for root, dirs, files in os.walk(self.folder):
            for doc in files:
                file_list.append(doc)
        # extract data
        for doc in file_list:
            i = os.path.join(self.folder , doc)
            if self.extract_files(i, retrieve_files):
                backup_data(i, doc)
        if retrieve_files == None:
            message_exit("Aucun don n'est périmé.")

    def extract_files(self, doc, retrieve_files):
        """
        extracts wallets data from each readable file and adds it to retrieve_files.
        each found pubkey will be a ["key"] for retrieve_files dict and will contain all relevant wallets data.
        """
        if os.path.isfile(doc):
            try:
                pubkey, content  = self.open_data(doc)
            except:
                print ("ERREUR : le fichier {0} n'est pas lisible".format(doc))
                return False
            if self.is_date_passed(content):
                try:
                    retrieve_files[pubkey].append(content)
                except:
                    retrieve_files[pubkey] = list()
                    retrieve_files[pubkey].append(content)
                print ("Récupération du fichier : {0} périmé le {1}".format(doc, content["peremption_date"]))
                return True
            else:
                print("Fichier non récupéré, périmera le {1} : {0} .".format(doc, content["peremption_date"]))
                return False

    @coroutine
    async def retrieve(self):
        """
        retrieve money from perished tips
        """
        client = ClientInstance().client
        error = False
        retrieve_files = dict()
        if hasattr(self, "folder"):
            self.sort_folder(retrieve_files)
        elif hasattr(self, "file"):
            self.extract_files(self.file, retrieve_files)

        key_list = list()
        for pubkey in retrieve_files:
            key = self.get_pubkey(pubkey)
            key_list.append(key)
        for key, pubkey in zip(key_list, retrieve_files):
            if key is not None:
                print("Récupération des pourboires créée par la clef {0}...".format(pubkey))
                for content in retrieve_files[pubkey]:
                    ok = await self.retrieve_wallets(content, key)
                    if ok is False:
                        error = True
        if error:
            print("ERREUR : Des erreurs sont survenues, certains dons ont pu ne pas être récupérés. Relisez bien les logs.")
        # closing the client instance
        await client.close()

    def open_data(self, _file):
        """
        open encrypted data
        """
        data_file = open(_file, "r")
        signed_data = data_file.read()
        pubkey = re.search(re.compile(PUBKEY_PATTERN), signed_data).group()
        data = ascii_armor.AsciiArmor.parse(signed_data, sender_pubkeys=[pubkey])
        data_file.close()
        return pubkey, json.loads(data["message"]["content"])

    async def retrieve_wallets(self, content, key):
        """
        send money from wallets
        """
        try :
            encrypted_wallets = content["wallets"]
            retrieving_wallets = ascii_armor.AsciiArmor.parse(encrypted_wallets, key) 
            retrieving_wallets = json.loads(retrieving_wallets["message"]["content"])
        except:
            print("ERREUR : Des dons, périmés le {0}, n'ont pas pu être récupérés, probablement à cause d'une modification du fichier ou d'un fichier mal formaté.".format(content["peremption_date"]))
            return False
        # send transactions
        while len(retrieving_wallets) != 0:
            _range, issuers, key_list, amount, listinput = 0, list(), list(), 0, list()
            wallet_list = list()
            if len(retrieving_wallets) > MAX_ISSUERS:
                for n in range (0,MAX_ISSUERS):
                    wallet_list.append (retrieving_wallets[n])
            else :
                wallet_list = retrieving_wallets
            issuers_keys = self.get_key_list(wallet_list)
            #create lists : keys, issuers, amount
            _range, intermediatetransaction = 0, 0
            for a in range (0, len(issuers_keys)):
                # print(a, _range) #debug
                # get inputs from all issuers
                listinput_and_amount = await get_list_input_for_transaction(_range, issuers_keys[a].pubkey, len(listinput))
                # print each empty G1Don
                if listinput_and_amount[1] == 0:
                    print("Ce don est vide ou a été récupéré : {0}".format(issuers_keys[a].pubkey))
                else:
                    _range += 1
                    key_list.append(issuers_keys[a])
                    issuers.append(issuers_keys[a].pubkey)
                    listinput.extend(listinput_and_amount[0])
                    amount = amount + listinput_and_amount[1]
                    intermediatetransaction = listinput_and_amount[2]
                # send transactions (intermediate if necessary)
                if intermediatetransaction or (issuers_keys[a] == issuers_keys[-1] and amount !=0):
                    # if a wallet has a lot of sources, we want to use all sources in many transactions.
                    while amount != 0:
                        total_listinput_and_amount = [listinput, amount, intermediatetransaction]
                        #amount = [amount]
                        await generate_and_send_transaction(
                            key_list,
                            issuers,
                            [amount],
                            total_listinput_and_amount,
                            [key.pubkey],
                            COMMENT,
                        )
                        key_list, issuers = [issuers_keys[a]], [issuers_keys[a].pubkey]
                        listinput, amount, intermediatetransaction = await get_list_input_for_transaction(0, issuers_keys[a].pubkey, 0)
                        sleep(1)
                    _range, issuers, key_list, amount, listinput = 0, list(), list(), 0, list()
            # delete already used wallets
            if not len(retrieving_wallets) == 0:
                limit = min(len(retrieving_wallets), MAX_ISSUERS)
                for n in range (0, limit):
                    retrieving_wallets.pop(0)

class BackupRetrieve():
    """
    prepare the retrieval of backups.
    """
    def __init__(self, year):

        conf = CheckConfig()
        conf.check()

        year = self.check_year(year)
        self.orig = os.path.join(conf.backup_folder, year)
        self.dest = os.path.join(conf.folder, year)
        if not os.path.isdir(self.orig):
            message_exit("ERREUR : {0} n'a pas de dossier correspondant dans {1}.".format(year, self.orig))
        if os.path.exists(self.dest):
            try:
                shutil.rmtree(self.dest)
            except:
                message_exit("ERREUR : le dossier {0} n'a pas pu être supprimé. Supprimez-le et recommencez.".format(self.dest))
        shutil.copytree(self.orig, self.dest)
        print("Récupération des backups {0}".format(year))
        print("Dossier {0} copié dans {1}".format(self.orig, self.dest))
        print("Si le processus ne se termine pas, vous pouvez le relancer.")

    def __call__(self):
        return self

    def delete(self):
        try:
            shutil.rmtree(self.dest)
        except:
            message_exit("ERREUR : le dossier {0} n'a pas pu être supprimé. Supprimez-le manuellement.".format(self.dest))

    def check_year(self, year):
        if re.search(re.compile("^[0-9]{4}$"), year):
            return str(year + "/")
        message_exit("ERREUR : {0} n'est pas une année valide.".format(str(year)))



import datetime
import re
import os
import json
from tabulate import tabulate

from gdon.generator import Generator
from gdon.transfer import Transfer
from gdon.constants import DEFAULT_FOLDER, WALLETS_FOLDER, BACKUP_FOLDER, PDF_EXTENSION, SIGNED_EXTENSION, CONF_FILE, MAX_PAGES, NAME
from gdon.utils import CheckConfig, check_date
from gdon.silkaj.tools import coroutine, message_exit
from gdon.silkaj.auth import auth_by_scrypt
from gdon.silkaj.constants import G1_SYMBOL
from duniterpy.key import SigningKey, ascii_armor


# Generate and fill Tips
class Creating:
    """Generating, filling wallets"""

    def __init__(self, amount, given_date, pages=1):
        if pages > MAX_PAGES:
            message_exit("Erreur : Ğ1Don gère un maximum de 10 pages.")
        self.pages = pages
        self.amount = amount
        conf = CheckConfig()
        conf.check()
        self.folder, self.wallets_folder, self.backup_folder = conf.folder, conf.wallets_folder, conf.backup_folder
        self.output = datetime.datetime.now().strftime(
            "Ğ1Don_%Y-%m-%dT%Hh%M"
        )
        self.pdf_output = self.folder + self.output + PDF_EXTENSION
        self.signed_output = self.wallets_folder + self.output + SIGNED_EXTENSION
        if os.path.isfile(self.pdf_output) or os.path.isfile(self.signed_output):
            message_exit("ERREUR : {0}{1}/{2} existe déjà. Attendez une minute et recommencez.".format(self.output, PDF_EXTENSION, SIGNED_EXTENSION))
        self.wallets = []
        # Check date format
        self.date = check_date(given_date)
        self.key = auth_by_scrypt(None)

    def __call__(self):
        return self

    def create(self):
        """ Generating, filling wallets."""

        if self.key and self.pages <= MAX_PAGES:
            print ("\nCréation des Ğ1Dons...")
            Generator(self.pdf_output, self.wallets, self.amount, self.date, self.pages).generate()
            print ("Enregistrement des portefeuilles sur :\n{0}\n".format(self.pdf_output))
            if not self.confirmation():
                os.remove(self.pdf_output)
                message_exit("OK, à bientôt !")
            try:
                self.save_json()
                transfer = Transfer(self.amount, self.pages, self.wallets, self.date)
                if self.amount > 0:
                    print ("Transferts...")
                    transfer.transfer(self.key)
            except:
                self.transfer_error()

    def transfer_error(self):
        """
        transfer the newly created files to the backup folder.
        """
        os.rename(self.pdf_output, self.backup_folder + self.output + PDF_EXTENSION)
        os.rename(self.signed_output, self.backup_folder + self.output + SIGNED_EXTENSION) 
        print ("\nERREUR : Quelque chose s'est mal passé durant le transfert.\nLes fichiers {0}{1}/{2} ont été déplacés dans le dossier {3}.\nVérifiez que votre clef publique est la bonne.\nVérifiez que le noeud Duniter est disponible.\nVérifiez si le transfert a eu lieu.".format(self.output, PDF_EXTENSION, SIGNED_EXTENSION, self.backup_folder))

    def save_json(self):
        """
        write wallets data in an encrypted json file.
        """
        # create and encrypt wallet list
        wallet_data = [
            {
                "pubkey": w["pubkey"],
                "salt": w["salt"],
                "password": w["password"],
            }
            for w in self.wallets
        ]
        clear_json = json.dumps(wallet_data)
        encrypted_wallets = ascii_armor.AsciiArmor.create(clear_json, self.key.pubkey)

        # add date and creator info
        data = {
            "creator_pubkey": self.key.pubkey,
            "peremption_date": self.date.isoformat(),
            "wallets": encrypted_wallets,
        }
        data = json.dumps(data)
        # sign the whole document
        signed_data = ascii_armor.AsciiArmor.create(data, signing_keys=[self.key])
        with open(
            self.signed_output, "a"
        ) as f:
            f.write(signed_data)


    def confirmation(self):
        tx=list()
        tx.append(["Compte créateur", self.key.pubkey])
        tx.append(["Ğ1Dons", "Montant"])
        for i, w in enumerate(self.wallets):
            tx.append([w["pubkey"], str(self.amount) + " " + G1_SYMBOL])
        tx.append(["Total", str(self.amount * len(self.wallets)) + " " + G1_SYMBOL])

        print(
            "Provisionner les", NAME, ":\n" +
            tabulate(tx, tablefmt="fancy_grid"),
        )
        answer = input("\nEnvoyer la transaction ? (O/n) : ")
        if answer == "O" or answer == "o" or answer == "" :
            return True
        return False



"""
Copyright  2016-2019 Maël Azimi <m.a@moul.re>

Silkaj is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Silkaj is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with utils.silkaj. If not, see <https://www.gnu.org/licenses/>.
"""

#from click import command, argument, pass_context

from gdon.silkaj.network_tools import ClientInstance, HeadBlock
from gdon.silkaj.tools import CurrencySymbol, message_exit, coroutine
from gdon.silkaj.auth import auth_method
#from gdon.silkaj.wot import check_public_key
from duniterpy.api.bma import tx, blockchain
from duniterpy.documents.transaction import InputSource


#@command("balance", help="Get wallet balance")
#@argument("pubkeys", nargs=-1)
#@pass_context
@coroutine
async def cmd_amount(ctx, pubkeys):
    client = ClientInstance().client
    if not (
        ctx.obj["AUTH_SCRYPT"]
        or ctx.obj["AUTH_FILE"]
        or ctx.obj["AUTH_SEED"]
        or ctx.obj["AUTH_WIF"]
    ):
        if not pubkeys:
            message_exit("You should specify one or many pubkeys")
        for pubkey in pubkeys:
            pubkey = check_public_key(pubkey, True)
            if not pubkey:
                return
        total = [0, 0]
        for pubkey in pubkeys:
            value = await get_amount_from_pubkey(pubkey)
            await show_amount_from_pubkey(pubkey, value)
            total[0] += value[0]
            total[1] += value[1]
        if len(pubkeys) > 1:
            await show_amount_from_pubkey("Total", total)
    else:
        key = auth_method()
        pubkey = key.pubkey
        await show_amount_from_pubkey(pubkey, await get_amount_from_pubkey(pubkey))
    await client.close()


async def show_amount_from_pubkey(pubkey, value):
    totalAmountInput = value[0]
    amount = value[1]
    currency_symbol = await CurrencySymbol().symbol
    ud_value = await UDValue().ud_value
    average, monetary_mass = await get_average()
    if totalAmountInput - amount != 0:
        print("Blockchain:")
        print("-----------")
        print("Relative     =", round(amount / ud_value, 2), "UD", currency_symbol)
        print("Quantitative =", round(amount / 100, 2), currency_symbol + "\n")

        print("Pending Transaction:")
        print("--------------------")
        print(
            "Relative     =",
            round((totalAmountInput - amount) / ud_value, 2),
            "UD",
            currency_symbol,
        )
        print(
            "Quantitative =",
            round((totalAmountInput - amount) / 100, 2),
            currency_symbol + "\n",
        )

    print("Total amount of: " + pubkey)
    print("----------------------------------------------------------------")
    print(
        "Total Relative     =",
        round(totalAmountInput / ud_value, 2),
        "UD",
        currency_symbol,
    )
    print("Total Quantitative =", round(totalAmountInput / 100, 2), currency_symbol)
    print(
        "Total Relative to average money share =",
        round(totalAmountInput / average, 2),
        "× M/N",
    )
    print(
        "Total Relative to monetary mass       =",
        round((totalAmountInput / monetary_mass) * 100, 3),
        "% M" + "\n",
    )


async def get_average():
    head = await HeadBlock().head_block
    monetary_mass = head["monetaryMass"]
    members_count = head["membersCount"]
    average = monetary_mass / members_count
    return average, monetary_mass


async def get_amount_from_pubkey(pubkey):
    listinput, amount = await get_sources(pubkey)

    totalAmountInput = 0
    for input in listinput:
        totalAmountInput += amount_in_current_base(input)
    return totalAmountInput, amount


async def get_sources(pubkey):
    client = ClientInstance().client
    # Sources written into the blockchain
    sources = await client(tx.sources, pubkey)

    listinput = list()
    amount = 0
    for source in sources["sources"]:
        if source["conditions"] == "SIG(" + pubkey + ")":
            listinput.append(
                InputSource(
                    amount=source["amount"],
                    base=source["base"],
                    source=source["type"],
                    origin_id=source["identifier"],
                    index=source["noffset"],
                )
            )
            amount += amount_in_current_base(listinput[-1])

    # pending source
    history = await client(tx.pending, pubkey)
    history = history["history"]
    pendings = history["sending"] + history["receiving"] + history["pending"]

    # add pending output
    pending_sources = list()
    for pending in pendings:
        identifier = pending["hash"]
        for i, output in enumerate(pending["outputs"]):
            outputsplited = output.split(":")
            if outputsplited[2] == "SIG(" + pubkey + ")":
                inputgenerated = InputSource(
                    amount=int(outputsplited[0]),
                    base=int(outputsplited[1]),
                    source="T",
                    origin_id=identifier,
                    index=i,
                )
                if inputgenerated not in listinput:
                    listinput.append(inputgenerated)

        for input in pending["inputs"]:
            pending_sources.append(InputSource.from_inline(input))

    # remove input already used
    for input in pending_sources:
        if input in listinput:
            listinput.remove(input)
    return listinput, amount
        

class UDValue(object):
    __instance = None

    def __new__(cls):
        if UDValue.__instance is None:
            UDValue.__instance = object.__new__(cls)
        return UDValue.__instance

    def __init__(self):
        self.ud_value = self.get_ud_value()

    async def get_ud_value(self):
        client = ClientInstance().client
        blockswithud = await client(blockchain.ud)
        NBlastUDblock = blockswithud["result"]["blocks"][-1]
        lastUDblock = await client(blockchain.block, NBlastUDblock)
        return lastUDblock["dividend"] * 10 ** lastUDblock["unitbase"]


def amount_in_current_base(source):
    """
    Get amount in current base from input or output source
    """
    return source.amount * 10 ** source.base



"""
Copyright  2016-2019 Maël Azimi <m.a@moul.re>

Silkaj is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Silkaj is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with utils.silkaj. If not, see <https://www.gnu.org/licenses/>.
"""

from re import compile, search
import math
from time import sleep
#from click import command, option, FloatRange

#from tabulate import tabulate
from gdon.silkaj.network_tools import ClientInstance, HeadBlock
from gdon.silkaj.crypto_tools import check_public_key
from gdon.silkaj.tools import message_exit, CurrencySymbol, coroutine
from gdon.silkaj.auth import auth_method
#from gdon.silkaj.wot import is_member
from gdon.silkaj.money import (
    get_sources,
    get_amount_from_pubkey,
    UDValue,
    amount_in_current_base,
)
from gdon.silkaj.constants import SOURCES_PER_TX

from duniterpy.api.bma.tx import process
from duniterpy.documents import BlockUID, Transaction
from duniterpy.documents.transaction import OutputSource, Unlock, SIGParameter

# max size for tx doc is 100 lines. Formula for accepted field numbers is : (2 * IU + 2 * IS + O) <= ( MAX_LINES - FIX_LINES)
# with IU = inputs/unlocks ; IS = Issuers/Signatures ; O = Outouts.
MAX_LINES_IN_TX_DOC = 100
# 2 lines are necessary, and we block 1 more for the comment
FIX_LINES = 3
# assuming there is only 1 issuer and 2 outputs, max inputs is 46
MAX_INPUTS_PER_TX = 46
# assuming there is 1 issuer and 1 input, max outputs is 93.
MAX_OUTPUTS = 93


"""
#@command("tx", help="Send transaction")
#@option(
    "--amount",
    help="Quantitative value(s) : <value_1>:<value_2>:... If only one amount, it will be sent to all recipients.",
)
#@option(
    "--amountUD",
    help="Relative value(s) : <UDvalue_1>:<UDvalue_2>... If only one amount, it will be sent to all recipients.",
)
#@option("--allSources", is_flag=True, help="Send all sources")
#@option(
    "--output",
    required=True,
    help="Pubkey(s)’ recipients + optional checksum: <pubkey>[!checksum]:[<pubkey>[!checksum]]",
)
#@option("--comment", default="", help="Comment")
#@option(
    "--outputBackChange",
    help="Pubkey recipient to send the rest of the transaction: <pubkey[!checksum]>",
)
#@option("--yes", "-y", is_flag=True, help="Assume yes. Do not prompt confirmation")
"""

@coroutine
async def send_transaction(
    amount, amountud, allsources, output, comment, outputbackchange, yes
):
    """
    Main function
    """
    tx_amount = await transaction_amount(amount, amountud, allsources)
    key = auth_method()
    issuer_pubkey = key.pubkey

    pubkey_amount = await get_amount_from_pubkey(issuer_pubkey)
    outputAddresses = output.split(":")
    totalAmount = total_amount(tx_amount, outputAddresses)
    if allsources:
        tx_amount = list(pubkey_amount[0])
    check_transaction_values(
        comment,
        outputAddresses,
        outputbackchange,
        pubkey_amount[0] < totalAmount,
        issuer_pubkey,
    )

    if (
        yes
        or input(
            tabulate(
                await transaction_confirmation(
                    issuer_pubkey,
                    pubkey_amount[0],
                    tx_amount,
                    totalAmount,
                    outputAddresses,
                    outputbackchange,
                    comment,
                ),
                tablefmt="fancy_grid",
            )
            + "\nDo you confirm sending this transaction? [yes/no]: "
        )
        == "yes"
    ):
        await handle_intermediaries_transactions(
            key,
            issuer_pubkey,
            tx_amount,
            totalAmount,
            outputAddresses,
            comment,
            outputbackchange,
        )


def total_amount(tx_amount, outputAddresses):
    """
    Return total amount of transaction
    """
    if len(tx_amount) != len(outputAddresses) and len(tx_amount) != 1:
        message_exit(
            "--output list and --amount or --amoundUD list are not of same length"
        )
    if len(tx_amount) == 1:
        if isinstance(outputAddresses, str):
            total = tx_amount[0]
        else :
            total = tx_amount[0] * len(outputAddresses)
    if len(tx_amount) > 1:
        total = sum(tx_amount)
    return total


async def transaction_amount(amount, amountUD, allSources):
    """
    Check command line interface amount option
    Return transaction amounts (list)
    """
    if not (amount or amountUD or allSources):
        message_exit("--amount nor --amountUD nor --allSources is set")
    amounts = list()
    if amount:
        for amount in amount.split(":"):
            amount = round(float(amount) * 100)	# , -2) ## CAUTION ! "-2" is for testing purpose, on GTest. Change to `round ( float(amount) *100 )`
            amounts.append(amount)
        return amounts

    if amountUD:
        for amount in amountUD.split(":"):
            amount = round(
                float(amount) * await UDValue().ud_value)	# , -2)
#            )  # DEBUG : ", -1"  est pour GTest, à enlever ! # ajouter la gestion de UnitBase ?
            amounts.append(amount)
        return amounts


def check_transaction_values(
    comment, outputAddresses, outputBackChange, enough_source, issuer_pubkey
):
    checkComment(comment)
    for outputAddress in outputAddresses:
        if check_public_key(outputAddress, True) is False:
            message_exit(outputAddress)
    if outputBackChange:
        outputBackChange = check_public_key(outputBackChange, True)
        if check_public_key(outputBackChange, True) is False:
            message_exit(outputBackChange)
    if enough_source:
        message_exit(
            issuer_pubkey + " pubkey doesn’t have enough money for this transaction."
        )


async def display_amount(tx, message, amount, currency_symbol):
    """
    For transaction_confirmation,
    Displays an amount in unit and relative reference.
    """
    amount_UD = round((amount / await UDValue().ud_value), 4)
    tx.append(
        [
            message + " (unit | relative)",
            "{unit_amount} {currency_symbol}  ||  {UD_amount} UD".format(
                unit_amount=str(amount / 100),
                currency_symbol=currency_symbol,
                UD_amount=str(amount_UD),
            ),
        ]
    )


async def display_pubkey(tx, message, pubkey):
    """
    For transaction_confirmation,
    Displays a pubkey and the eventually associated id.
    """
    tx.append([message + " (pubkey)", pubkey])
    id = await is_member(pubkey)
    if id:
        tx.append([message + " (id)", id["uid"]])


async def display_output_and_amount(tx, outputAddresses, tx_amount, currency_symbol):
    """
    For transaction_confirmation,
    Displays the receiver(s) and the amount(s) of one or many output(s).
    """
    if len(tx_amount) == 1:
        for outputAddress in outputAddresses:
            await display_pubkey(tx, "to", outputAddress)
            await display_amount(tx, "amount", tx_amount[0], currency_symbol)
    if len(tx_amount) > 1:
        c = 0
        while c < len(outputAddresses):
            await display_pubkey(tx, "to", outputAddresses[c])
            await display_amount(tx, "amount", tx_amount[c], currency_symbol)
            c += 1


async def transaction_confirmation(
    issuer_pubkey,
    pubkey_amount,
    tx_amount,
    totalAmount,
    outputAddresses,
    outputBackChange,
    comment,
):
    """
    Generate transaction confirmation
    """

    currency_symbol = await CurrencySymbol().symbol
    tx = list()
    tx.append(
        ["pubkey’s balance before tx", str(pubkey_amount / 100) + " " + currency_symbol]
    )

    await display_amount(tx, "total amount", totalAmount, currency_symbol)

    tx.append(
        [
            "pubkey’s balance after tx",
            str(((pubkey_amount - totalAmount) / 100)) + " " + currency_symbol,
        ]
    )

    await display_pubkey(tx, "from", issuer_pubkey)
    await display_output_and_amount(tx, outputAddresses, tx_amount, currency_symbol)
    if outputBackChange:
        await display_pubkey(tx, "Backchange", outputBackChange)

    tx.append(["comment", comment])
    return tx


async def get_list_input_for_transaction(pubkey, TXamount, outputs_number):
    listinput, amount = await get_sources(pubkey)

    # check max inputs. For now we deal only with one issuer
    maxInputsNumber = max_inputs_number(outputs_number, 1)
    # generate final list source
    listinputfinal = []
    totalAmountInput = 0
    intermediatetransaction = False
    for input in listinput:
        listinputfinal.append(input)
        totalAmountInput += amount_in_current_base(input)
        TXamount -= amount_in_current_base(input)
        # if too much sources, it's an intermediate transaction
        if len(listinputfinal) >= maxInputsNumber and not len(listinputfinal) == 1:
            intermediatetransaction = True
        if (len(listinputfinal) >= MAX_INPUTS_PER_TX) or (TXamount <= 0):
            break
    if TXamount > 0 and not intermediatetransaction:
        message_exit("Error: you don't have enough money")
    return listinputfinal, totalAmountInput, intermediatetransaction


def max_inputs_number(outputs_number, issuers_number=1, comment=True):
    """
    Returns the maximum number of inputs.
    This function does not take care of backchange line.
    Formula is Inputs_Unlock number <= (MAX_LINES - FIX_LINES - Comment - Output_number - 2* Issuer_Signature_number)/2
    """
    if comment == True:
        comment = 1
    return (
        MAX_LINES_IN_TX_DOC
        - FIX_LINES
        - comment
        - (2 * issuers_number)
        - outputs_number
    ) / 2



# Créer une fonction handle_transactions_with_conditions
# qui gèrerait l'ajout de conditions spécifiques pour G1pourboire :
# SIG(0) || SIG(sender_pubkey) && CSV(time)
async def handle_intermediaries_transactions(
    key,
    issuers,
    tx_amount,
    totalAmount,
    outputAddresses,
    Comment="",
    OutputbackChange=None,
):
    client = ClientInstance().client
    while True:
        listinput_and_amount = await get_list_input_for_transaction(
            issuers, totalAmount, len(outputAddresses)
        )

        intermediatetransaction = listinput_and_amount[2]

        if intermediatetransaction:
            totalAmountInput = list()
            totalAmountInput.append (listinput_and_amount[1])
            await generate_and_send_transaction(
                key,
                issuers,
                totalAmount,
                totalAmountInput,
                listinput_and_amount,
                issuers,
                "Change operation",
            )
            sleep(1)  # wait 1 second before sending a new transaction
        else:
            await generate_and_send_transaction(
                key,
                issuers,
                totalAmount,
                tx_amount,
                listinput_and_amount,
                outputAddresses,
                Comment,
                OutputbackChange,
            )
            await client.close()
            break
        #return 0

async def handle_transactions_with_delay(
    key,
    issuers,
    tx_amount,
    totalAmount,
    outputAddresses,
    delay,
    Comment="",
    OutputbackChange=None,
):
    client = ClientInstance().client
    while True:
        listinput_and_amount = await get_list_input_for_transaction(
            issuers, totalAmount
        )

        intermediatetransaction = listinput_and_amount[2]

        if intermediatetransaction:
            totalAmountInput = list()
            totalAmountInput.append (listinput_and_amount[1])
            await generate_and_send_transaction(
                key,
                issuers,
                totalAmount,
                totalAmountInput,
                listinput_and_amount,
                issuers,
                "Change operation",
            )
            sleep(1)  # wait 1 second before sending a new transaction
        else:
            await generate_and_send_transaction_with_delay(
                key,
                issuers,
                totalAmount,
                tx_amount,
                listinput_and_amount,
                outputAddresses,
                delay,
                Comment,
                OutputbackChange,
            )
            await client.close()
            break


async def generate_and_send_transaction(
    key,
    issuers,
    totalAmount,
    tx_amount,
    listinput_and_amount,
    outputAddresses,
    Comment,
    OutputbackChange=None,
):
    """
    Display sent transaction
    Generate, sign, and send transaction document
    """
    intermediate_tx = listinput_and_amount[2]
    if intermediate_tx:
        print("Generate Change Transaction")
    else:
        print("Generate Transaction:")
    print("   - From:    " + issuers)
    if len(tx_amount) == 1:
        if isinstance(outputAddresses, str):
            display_sent_tx(outputAddresses, tx_amount[0])
        else:
            for outputAddress in outputAddresses:
                display_sent_tx(outputAddress, tx_amount[0])
    else:
        c = 0
        while c < len(outputAddresses):
            display_sent_tx(outputAddresses[c], tx_amount[c])
            c += 1
    client = ClientInstance().client
    transaction = await generate_transaction_document(
        issuers,
        totalAmount,
        tx_amount,
        listinput_and_amount,
        outputAddresses,
        Comment,
        OutputbackChange,
    )
    transaction.sign([key])
    response = await client(process, transaction.signed_raw())
    if response.status == 200:
        print("Transaction successfully sent.")
    else:
        message_exit(
            "Error while publishing transaction: {0}".format(await response.text())
        )

async def generate_and_send_transaction_with_delay(
    key,
    issuers,
    totalAmount,
    tx_amount,
    listinput_and_amount,
    outputAddresses,
    delay,
    Comment,
    OutputbackChange=None,
):
    """
    Display sent transaction
    Generate, sign, and send transaction document
    """
    intermediate_tx = listinput_and_amount[2]
    if intermediate_tx:
        print("Generate Change Transaction")
    else:
        print("Generate Transaction:")
    print("   - From:    " + issuers)
    if len(tx_amount) == 1:
        if isinstance(outputAddresses, str):
            display_sent_tx(outputAddresses, tx_amount[0])
        else:
            for outputAddress in outputAddresses:
                display_sent_tx(outputAddress, tx_amount[0])
    else:
        c = 0
        while c < len(outputAddresses):
            display_sent_tx(outputAddresses[c], tx_amount[c])
            c += 1
    client = ClientInstance().client
    transaction = await generate_transaction_document_with_delay(
        issuers,
        totalAmount,
        tx_amount,
        listinput_and_amount,
        outputAddresses,
        delay,
        Comment,
        OutputbackChange,
    )
    transaction.sign([key])
    response = await client(process, transaction.signed_raw())
    if response.status == 200:
        print("Transaction successfully sent.")
    else:
        message_exit(
            "Error while publishing transaction: {0}".format(await response.text())
        )


def display_sent_tx(outputAddress, amount):
    print("   - To:     ", outputAddress, "\n   - Amount: ", int(amount) / 100)

async def generate_transaction_document(
    issuers,
    totalAmount,
    tx_amount,
    listinput_and_amount,
    outputAddresses,
    Comment="",
    OutputbackChange=None,
):

    listinput = listinput_and_amount[0]
    totalAmountInput = listinput_and_amount[1]

    head_block = await HeadBlock().head_block
    currency_name = head_block["currency"]
    blockstamp_current = BlockUID(head_block["number"], head_block["hash"])
    curentUnitBase = head_block["unitbase"]

    if not OutputbackChange:
        OutputbackChange = issuers

    # if it's not a foreign exchange transaction, we remove units after 2 digits after the decimal point.
    if issuers not in outputAddresses:
        totalAmount = (totalAmount // 10 ** curentUnitBase) * 10 ** curentUnitBase

    # Generate output
    ################
    listoutput = []
    if len(tx_amount) == 1:
    # Outputs to himself (change operation)
        if isinstance(outputAddresses, str):
            generate_output(listoutput, curentUnitBase, tx_amount[0], outputAddresses)
    # Outputs to receiver (if not himself)
        else:
            for outputAddress in outputAddresses:
                generate_output(listoutput, curentUnitBase, tx_amount[0], outputAddress)
    else:
        c = 0
        while c < len(outputAddresses):
            generate_output(
                listoutput, curentUnitBase, tx_amount[c], outputAddresses[c]
            )
            c += 1
    # Outputs to himself
    rest = totalAmountInput - totalAmount
    generate_output(listoutput, curentUnitBase, rest, OutputbackChange)

    # Unlocks
    unlocks = generate_unlocks(listinput)

    # Generate transaction document
    ##############################

    return Transaction(
        version=10,
        currency=currency_name,
        blockstamp=blockstamp_current,
        locktime=0,
        issuers=[issuers],
        inputs=listinput,
        unlocks=unlocks,
        outputs=listoutput,
        comment=Comment,
        signatures=[],
    )

async def generate_transaction_document_with_delay(
    issuers,
    totalAmount,
    tx_amount,
    listinput_and_amount,
    outputAddresses,
    delay,
    Comment="",
    OutputbackChange=None,
):

    listinput = listinput_and_amount[0]
    totalAmountInput = listinput_and_amount[1]

    head_block = await HeadBlock().head_block
    currency_name = head_block["currency"]
    blockstamp_current = BlockUID(head_block["number"], head_block["hash"])
    curentUnitBase = head_block["unitbase"]

    if not OutputbackChange:
        OutputbackChange = issuers

    # if it's not a foreign exchange transaction, we remove units after 2 digits after the decimal point.
    if issuers not in outputAddresses:
        totalAmount = (totalAmount // 10 ** curentUnitBase) * 10 ** curentUnitBase

    # Generate output
    ################
    listoutput = []
    if len(tx_amount) == 1:
    # Outputs to himself (change operation)
        if isinstance(outputAddresses, str):
            generate_output(listoutput, curentUnitBase, tx_amount[0], outputAddresses)
    # Outputs to receiver (if not himself)
        else:
            for outputAddress in outputAddresses:
                generate_output_with_delay(listoutput, curentUnitBase, tx_amount[0], outputAddress, delay, issuers)
    else:
        c = 0
        while c < len(outputAddresses):
            generate_output_with_delay(
                listoutput, curentUnitBase, tx_amount[c], outputAddresses[c], delay, issuers
            )
            c += 1
    # Outputs to himself
    rest = totalAmountInput - totalAmount
    generate_output(listoutput, curentUnitBase, rest, OutputbackChange)

    # Unlocks
    unlocks = generate_unlocks(listinput)

    # Generate transaction document
    ##############################

    return Transaction(
        version=10,
        currency=currency_name,
        blockstamp=blockstamp_current,
        locktime=0,
        issuers=[issuers],
        inputs=listinput,
        unlocks=unlocks,
        outputs=listoutput,
        comment=Comment,
        signatures=[],
    )


def generate_unlocks(listinput):
    unlocks = list()
    for i in range(0, len(listinput)):
        unlocks.append(Unlock(index=i, parameters=[SIGParameter(0)]))
    return unlocks


def generate_output(listoutput, unitbase, rest, recipient_address):
    while rest > 0:
        outputAmount = truncBase(rest, unitbase)
        rest -= outputAmount
        if outputAmount > 0:
            outputAmount = int(outputAmount / math.pow(10, unitbase))
            listoutput.append(
                OutputSource(
                    amount=str(outputAmount),
                    base=unitbase,
                    condition="SIG({0})".format(recipient_address),
                )
            )
        unitbase = unitbase - 1


def generate_output_with_delay(listoutput, unitbase, rest, recipient_address, delay, sender_address):
    while rest > 0:
        outputAmount = truncBase(rest, unitbase)
        rest -= outputAmount
        if outputAmount > 0:
            outputAmount = int(outputAmount / math.pow(10, unitbase))
            listoutput.append(
                OutputSource(
                    amount=str(outputAmount),
                    base=unitbase,
                    condition="(SIG({recipient}) || (SIG({sender}) && CSV({delay})))".format(recipient=recipient_address, sender=sender_address, delay=delay),
                )
            )

        unitbase = unitbase - 1


def checkComment(Comment):
    if len(Comment) > 255:
        message_exit("Error: Comment is too long")
    regex = compile(
        "^[0-9a-zA-Z\ \-\_\:\/\;\*\[\]\(\)\?\!\^\+\=\#@\&\~\#\{\}\|\\\<\>\%\.]*$"
    )
    if not search(regex, Comment):
        message_exit("Error: the format of the comment is invalid")


def truncBase(amount, base):
    pow = math.pow(10, base)
    if amount < pow:
        return 0
    return math.trunc(amount / pow) * pow



"""
Copyright  2016-2019 Maël Azimi <m.a@moul.re>

Silkaj is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Silkaj is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with utils.silkaj. If not, see <https://www.gnu.org/licenses/>.
"""

from nacl import encoding, hash
from re import compile, search


def check_public_key(pubkey, display_error):
    """
    Check public key format
    Check pubkey checksum which could be append after the pubkey
    If check pass: return pubkey
    """
    regex = compile("^[1-9A-HJ-NP-Za-km-z]{43,44}$")
    regex_checksum = compile(
        "^[1-9A-HJ-NP-Za-km-z]{43,44}" + "![1-9A-HJ-NP-Za-km-z]{3}$"
    )
    if search(regex, pubkey):
        return pubkey
    elif search(regex_checksum, pubkey):
        pubkey, checksum = pubkey.split("!")
        pubkey_byte = b58_decode(pubkey)
        checksum_calculed = b58_encode(
            hash.sha256(
                hash.sha256(pubkey_byte, encoding.RawEncoder), encoding.RawEncoder
            )
        )[:3]
        if checksum_calculed == checksum:
            return pubkey
        else:
            print("Error: bad checksum for following public key:")
            return False

    elif display_error:
        print("Error: bad format for following public key:", pubkey)
    return False


b58_digits = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"


def b58_encode(b):

    # Convert big-endian bytes to integer
    n = int("0x0" + encoding.HexEncoder.encode(b).decode("utf8"), 16)

    # Divide that integer into base58
    res = []
    while n > 0:
        n, r = divmod(n, 58)
        res.append(b58_digits[r])
    res = "".join(res[::-1])

    # Encode leading zeros as base58 zeros
    czero = 0
    pad = 0
    for c in b:
        if c == czero:
            pad += 1
        else:
            break
    return b58_digits[0] * pad + res


def b58_decode(s):
    if not s:
        return b""

    # Convert the string to an integer
    n = 0
    for c in s:
        n *= 58
        if c not in b58_digits:
            raise InvalidBase58Error(
                "Character %r is not a " + "valid base58 character" % c
            )
        digit = b58_digits.index(c)
        n += digit

    # Convert the integer to bytes
    h = "%x" % n
    if len(h) % 2:
        h = "0" + h
    res = encoding.HexEncoder.decode(h.encode("utf8"))

    # Add padding back.
    pad = 0
    for c in s[:-1]:
        if c == b58_digits[0]:
            pad += 1
        else:
            break
    return b"\x00" * pad + res



SUGGESTED_AMOUNT = 10
MAX_AMOUNT = 1000000000
SUGGESTED_DEV_PERCENT = 5
DATE_FORMAT = "[0-9]{2}/[0-9]{2}/[0-9]{4}"
PUBKEY_FORMAT = "[1-9A-HJ-NP-Za-km-z]{43,44}"
DEFAULT_FOLDER = "~/Documents/G1dons/"
WALLETS_FOLDER = "Recuperation_ne_pas_supprimer/"
BACKUP_FOLDER = ".backup/"
PDF_BACKUP_FOLDER = "pdf/"
PDF_EXTENSION = ".pdf"
SIGNED_EXTENSION = ".json.signed"
CONF_FILE = "~/.config/gdon.conf"
MAX_PAGES = 10
MAX_ISSUERS = 18
COMMENT = "G1Don"
NAME = "Ğ1Dons"
HTTP_PORT = 80



# Generate a pseudo diceware password
# DO NOT USE IT FOR STRONG CRYPTOGRAPHIC PURPOSE!

from random import choice
import sys
import os

# set the wordlist path
wordlist = "objects/wordlist.txt"

parent_dir = os.path.abspath(os.path.dirname(__file__))
wordlist_path = os.path.join(parent_dir, wordlist)

def diceware(N, wordlist=wordlist_path, separator="", camelcase=True):
	with open(wordlist, "r") as f:
		words = f.read().split('\n')
		out = []
		for i in range(N):
			word = choice(words)
			if camelcase:
				word = word.capitalize()
			out.append(word)
		return separator.join(out)



import os
import json
import datetime

from duniterpy.key import SigningKey, ascii_armor

from gdon.diceware import diceware
from gdon.silkaj.money import get_amount_from_pubkey
from gdon.silkaj.tx import handle_intermediaries_transactions
from gdon.silkaj.tools import coroutine
from gdon.constants import COMMENT

class Transfer:
    """Create a transfer wallet, send amount to tips"""

    def __init__(self, amount, pages, wallets, peremption_date):
        self.wallets = wallets
        self.pages = pages
        self.Tips_Amount = amount
        self.date = peremption_date


    @coroutine
    async def transfer(self, key):
        """
        sends tips.
        """
        total_amount = int(100 * self.Tips_Amount * 6 * self.pages)
        # create outputs list
        outputAddresses = []
        for w in self.wallets:
            outputAddresses.append (w["pubkey"])
        # send to tips
        outputbackchange = None
        amount_tx = int (self.Tips_Amount * 100)
        tx_amount = list()
        tx_amount.append(amount_tx)
        issuers = key.pubkey
        await handle_intermediaries_transactions(
            key,
            issuers,
            tx_amount,
            total_amount,
            outputAddresses,
            COMMENT,
            outputbackchange,
        )




from PIL import Image, ImageFont, ImageDraw
from reportlab.pdfgen import canvas
from reportlab.lib.units import cm
from duniterpy.key import SigningKey, ascii_armor
from gdon.diceware import diceware
from gdon.silkaj.constants import G1_SYMBOL

import os
import json
import qrcode
import datetime

recto_file = "objects/recto.png"
verso_file = "objects/verso.png"
typo_file = "objects/Roboto-Medium.ttf"

class Generator:
    """Generate some wallets, create the pdf to print"""

    def __init__(self, output, wallets, amount, date, pages=1):
        self.pages = pages
        self.output = output
        self.wallets = wallets
        self.amount = amount
        self.date = date
        self.c = None

        # set objects dir
        par_dir = os.path.abspath(os.path.dirname(__file__))
	#wordlist_path = os.path.join(parent_dir, wordlist)
        self.recto = os.path.join(par_dir, recto_file)
        self.verso = os.path.join(par_dir, verso_file)
        self.typo = os.path.join(par_dir, typo_file)

    def generate(self):
        """
        generate tips wallets ; then create the pdf file
        """
        for i in range(self.pages):
            for j in range(6):  # 6 wallets per pages
                self.new_wallet()

        self.make_pdf()

    def new_wallet(self):
        """
        generate one wallet with the associated images
        """
        # Generating credentials
        salt = diceware(3, separator="-", camelcase=False)
        password = diceware(3, separator="-", camelcase=False)
        # Generating key from credentials
        key = SigningKey.from_credentials(salt, password)
        # Url to redirect to in the public QR code
        account_url = "https://demo.cesium.app/#/app/wot/tx/" + key.pubkey + "/"
        # Generating wif data
        key.save_wif_file("privatekey.wif")
        wif_data = open("privatekey.wif").readlines()[-1].split(": ")[1]
        #try add a \n after wif
#        wif_data.join(["\n"])
        os.remove("privatekey.wif")
        # Creating the QR codes
        qr_pub = qrcode.make(account_url)
        qr_priv = qrcode.make(wif_data)
        # Open images
        recto = Image.open(self.recto)
        verso = Image.open(self.verso)
        # Pasting QR codes
        recto.paste(qr_pub.resize((200, 200)), (435, 15))  # 36))
        verso.paste(qr_priv.resize((180, 180)), (580, 6))
        # Setting font
#        font = ImageFont.truetype("objects/Roboto-Medium.ttf", 18)
        font = ImageFont.truetype(self.typo, 18)
        # Writing amount
        if self.amount > 0:
            draw = ImageDraw.Draw(recto)
            txt = "\n".join([str(self.amount) + G1_SYMBOL])
            draw.text(
                (1090, 150),
                txt,
                (0, 0, 0),
                font=ImageFont.truetype(self.typo, 30),
            )
        # Writing date
        write_date = "{0}/{1}/{2}".format(self.date.day, self.date.month, self.date.year)
        draw = ImageDraw.Draw(recto)
        draw.text(
            (665, 230),
            write_date,
            (0, 0, 0),
            font=font,
        )
        # Writing public key
        draw = ImageDraw.Draw(recto)
        txt = "\n".join([key.pubkey[i : i + 10] for i in range(0, len(key.pubkey), 10)])
        draw.text((665, 80), txt, (0, 0, 0), font=font)
        # Writing private keys
        draw = ImageDraw.Draw(verso)
        draw.text((570, 230), "ID : " + salt, (0, 0, 0), font=font)
        draw.text((570, 250), "PW : " + password, (0, 0, 0), font=font)
        # Add data to wallets
        self.wallets.append(
            {
                "salt": salt,
                "password": password,
                "pubkey": key.pubkey,
                "date": 0,
                "recto": recto,
                "verso": verso,
                "transfer": "false",
            }
        )

    def gen_page(self, wallets):
        """
        generate a pdf page from a wallet list with relevant images.
        """
        # Create a new canvas
        if self.c == None:
            self.c = canvas.Canvas(self.output)
        # Size of wallets. You may not edit those values until you know what you do
        width = 19 * cm
        height = width * 302 / 1270
        # Print recto
        for i, w in enumerate(wallets):
            self.c.drawInlineImage(
                w["recto"], (21 * cm - width) / 2, 24.2 * cm - i * height, width, height
            )
        self.c.showPage()
        # And verso
        for i, w in enumerate(wallets):
            self.c.drawInlineImage(
                w["verso"], (21 * cm - width) / 2, 24.2 * cm - i * height, width, height
            )
        self.c.showPage()

    def make_pdf(self):
        def chunks(l, n):
            """Yield successive n-sized chunks from l."""
            for i in range(0, len(l), n):
                yield l[i : i + n]

        # Create pages
        for wallets in chunks(self.wallets, 6):
            self.gen_page(wallets)
        self.c.save()



"""
Copyright  2016-2019 Maël Azimi <m.a@moul.re>

Silkaj is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Silkaj is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with utils.silkaj. If not, see <https://www.gnu.org/licenses/>.
"""

from datetime import datetime
from sys import exit
from asyncio import get_event_loop
from functools import update_wrapper

from gdon.silkaj.constants import G1_SYMBOL, GTEST_SYMBOL
from gdon.silkaj.blockchain_tools import BlockchainParams


def convert_time(timestamp, kind):
    ts = int(timestamp)
    date = "%Y-%m-%d"
    hour = "%H:%M"
    second = ":%S"
    if kind == "all":
        pattern = date + " " + hour + second
    elif kind == "date":
        pattern = date
    elif kind == "hour":
        pattern = hour
        if ts >= 3600:
            pattern += second
    return datetime.fromtimestamp(ts).strftime(pattern)


class CurrencySymbol(object):
    __instance = None

    def __new__(cls):
        if CurrencySymbol.__instance is None:
            CurrencySymbol.__instance = object.__new__(cls)
        return CurrencySymbol.__instance

    def __init__(self):
        self.symbol = self.get_symbol()

    async def get_symbol(self):
        params = await BlockchainParams().params
        if params["currency"] == "g1":
            return G1_SYMBOL
        elif params["currency"] == "g1-test":
            return GTEST_SYMBOL


def message_exit(message):
    print(message)
    exit(1)


def coroutine(f):
    def wrapper(*args, **kwargs):
        loop = get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))

    return update_wrapper(wrapper, f)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gdon",
    version="0.0.7",
    author="matograine",
    author_email="matograine@zaclys.net",
    description="gdon helps creating paper tips for the Ğ1 libre crypto-currency ; and retrieve them after expiry.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    download_url="https://git.duniter.org/matograine/g1pourboire/-/tags/0.0.7",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "Natural Language :: French",
    ],
    python_requires='>=3.6',
    install_requires=[
        # Silkaj packages
        "duniterpy==0.56",
        "ipaddress",
        "texttable",
        "tabulate",
        "pynacl",
        # G1don specific packages
        "Pillow>=6.0.0",
        "qrcode>=6.1",
        "reportlab>=3.5.23",
    ],
    scripts=["bin/gdon"],
)




"""
Copyright  2016-2019 Maël Azimi <m.a@moul.re>

Silkaj is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Silkaj is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with utils.silkaj. If not, see <https://www.gnu.org/licenses/>.
"""

from gdon.silkaj.network_tools import ClientInstance
from duniterpy.api.bma import blockchain


class BlockchainParams(object):
    __instance = None

    def __new__(cls):
        if BlockchainParams.__instance is None:
            BlockchainParams.__instance = object.__new__(cls)
        return BlockchainParams.__instance

    def __init__(self):
        self.params = self.get_params()

    async def get_params(self):
        client = ClientInstance().client
        return await client(blockchain.parameters)



"""
Copyright  2016-2019 Maël Azimi <m.a@moul.re>

Silkaj is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Silkaj is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with utils.silkaj. If not, see <https://www.gnu.org/licenses/>.
"""

SILKAJ_VERSION = "0.7.3"
G1_SYMBOL = "Ğ1"
GTEST_SYMBOL = "ĞTest"
G1_DEFAULT_ENDPOINT = "g1.duniter.org", "443"
G1_TEST_DEFAULT_ENDPOINT = "ts.gt.librelois.fr", "443"
CONNECTION_TIMEOUT = 10
ASYNC_SLEEP = 0.1
PUBKEY_PATTERN = "[1-9A-HJ-NP-Za-km-z]{43,44}"
SOURCES_PER_TX = 30
